#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.Output.Pipe import BGPPipe


class StdOutPipe(BGPPipe):

    def __init__(self):
        # No initialization required
        pass

    def output(self, output):
        print(output)



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.Output.Filter import BGPFilter


class MessageSubTypeFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, message):
        try:
            if message.type is not BGPStatics.MESSAGE_TYPE_UPDATE:
                # Skip messages that are no UPDATE messages
                return None

            for value in self.values:
                negated = False

                if value[0:1] == "~":
                    negated = True
                    value = value[1:]

                if value == "WITHDRAWAL":
                    check_value = BGPStatics.UPDATE_TYPE_WITHDRAWAL
                elif value == "ANNOUNCE":
                    check_value = BGPStatics.UPDATE_TYPE_ANNOUNCE
                elif value == "BOTH":
                    check_value = BGPStatics.UPDATE_TYPE_BOTH
                elif value == "NONE":
                    check_value = BGPStatics.UPDATE_TYPE_NONE
                else:
                    return None

                if not negated and message.subtype == check_value:
                    # Match on message type
                    return message

                if negated and message.subtype != check_value:
                    return message

            # Searched value was not found
            return None
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics


class BGPUpdateFlags:
    def __init__(self, payload):
        # Using bitmask to parse bit flags
        # First 4 bit of flag-byte are in use - the following 4 bit need to be ignored
        self.optional = True if (payload & BGPStatics.UPDATE_FLAG_OPTIONAL) is not 0 else False
        self.transitive = True if (payload & BGPStatics.UPDATE_FLAG_TRANSITIVE) is not 0 else False
        self.partial = True if (payload & BGPStatics.UPDATE_FLAG_PARTIAL) is not 0 else False
        self.length = True if (payload & BGPStatics.UPDATE_FLAG_LENGTH) is not 0 else False



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import struct

from pbgpp.BGP.Open.OptionalParameter import BGPOptionalParameter
from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPOptionalParameterCapability(BGPOptionalParameter):

    def __init__(self, payload):
        BGPOptionalParameter.__init__(self, payload)
        self.type = BGPStatics.OPEN_CAPABILITY
        self.capability_list = []
        self.__parse()

    def __parse(self):
        logger = logging.getLogger("pbgpp.BGPOptionalParameterCapability.__parse")

        self.parsed = True

        try:
            current_byte_position = 0
            continue_loop = True

            while continue_loop:
                capability_fields = struct.unpack("!BB", self.payload[current_byte_position:current_byte_position + 2])
                capability_type = capability_fields[0]
                capability_length = capability_fields[1]

                payload_start_marker = current_byte_position + 2
                payload_stop_marker = payload_start_marker + capability_length

                # Use factory pattern to determine the capability class type
                self.capability_list.append(BGPCapability.factory(capability_type, self.payload[payload_start_marker:payload_stop_marker]))

                current_byte_position += (2 + capability_length)
                if current_byte_position >= len(self.payload):
                    continue_loop = False

        except Exception as e:
            logger.warning("Unspecified error during packet parsing. Exception could be raised due to a malformed message.")
            self.error = True

    def json(self):
        capabilities = []

        for c in self.capability_list:
            capabilities.append(c.json())

        return {
            "type": self.type,
            "type_string": BGPTranslation.open_parameter(self.type),
            "capabilities": capabilities
        }


class BGPCapability:

    def __init__(self, payload):
        self.payload = payload
        self.type = None
        self.parsed = False
        self.error = None

    @staticmethod
    def factory(capability_type, payload):
        # Factory pattern for capabilities of capability parameter of OPEN messages
        if capability_type == BGPStatics.CAPABILITY_MULTIPROTOCOL_EXTENSIONS:
            from pbgpp.BGP.Open.Parameters.Capabilities.MultiprotocolExtensions import CapabilityMultiprotocolExtensions
            return CapabilityMultiprotocolExtensions(payload)

        if capability_type == BGPStatics.CAPABILITY_ROUTE_REFRESH:
            from pbgpp.BGP.Open.Parameters.Capabilities.RouteRefresh import CapabilityRouteRefresh
            return CapabilityRouteRefresh(payload)

        if capability_type == BGPStatics.CAPABILITY_ALTERNATIVE_ROUTE_REFRESH:
            from pbgpp.BGP.Open.Parameters.Capabilities.RouteRefresh import CapabilityRouteRefresh
            return CapabilityRouteRefresh(payload, legacy=True)

        if capability_type == BGPStatics.CAPABILITY_OUTBOUND_ROUTE_FILTERING:
            from pbgpp.BGP.Open.Parameters.Capabilities.OutboundRouteFiltering import CapabilityOutboundRouteFilter
            return CapabilityOutboundRouteFilter(payload)

        if capability_type == BGPStatics.CAPABILITY_MULTIPLE_ROUTES_TO_DESTINATION:
            from pbgpp.BGP.Open.Parameters.Capabilities.MultipleRoutesToDestination import CapabilityMultipleRoutesToDestination
            return CapabilityMultipleRoutesToDestination(payload)

        if capability_type == BGPStatics.CAPABILITY_EXTENDED_NEXT_HOP_ENCODING:
            from pbgpp.BGP.Open.Parameters.Capabilities.ExtendedNextHopEncoding import CapabilityExtendedNextHopEncoding
            return CapabilityExtendedNextHopEncoding(payload)

        if capability_type == BGPStatics.CAPABILITY_BGP_EXTENDED:
            from pbgpp.BGP.Open.Parameters.Capabilities.BGPExtended import CapabilityBGPExtended
            return CapabilityBGPExtended(payload)

        if capability_type == BGPStatics.CAPABILITY_GRACEFUL_RESTART:
            from pbgpp.BGP.Open.Parameters.Capabilities.GracefulRestart import CapabilityGracefulRestart
            return CapabilityGracefulRestart(payload)

        if capability_type == BGPStatics.CAPABILITY_SUPPORT_FOR_FOUR_OCTET_AS:
            from pbgpp.BGP.Open.Parameters.Capabilities.SupportForFourOctetAS import CapabilitySupportForFourOctetAS
            return CapabilitySupportForFourOctetAS(payload)

        if capability_type == BGPStatics.CAPABILITY_SUPPORT_FOR_DYNAMIC_CAPABILITY:
            from pbgpp.BGP.Open.Parameters.Capabilities.SupportForDynamicCapability import CapabilitySupportForDynamicCapability
            return CapabilitySupportForDynamicCapability(payload)

        if capability_type == BGPStatics.CAPABILITY_MULTISESSION_BGP:
            from pbgpp.BGP.Open.Parameters.Capabilities.MultisessionBGP import CapabilityMultisessionBGP
            return CapabilityMultisessionBGP(payload)

        if capability_type == BGPStatics.CAPABILITY_ADD_PATH:
            from pbgpp.BGP.Open.Parameters.Capabilities.AddPath import CapabilityAddPath
            return CapabilityAddPath(payload)

        if capability_type == BGPStatics.CAPABILITY_ENHANCED_ROUTE_REFRESH:
            from pbgpp.BGP.Open.Parameters.Capabilities.EnhancedRouteRefresh import CapabilityEnhancedRouteRefresh
            return CapabilityEnhancedRouteRefresh(payload)

        if capability_type == BGPStatics.CAPABILITY_LONG_LIVED_GRACEFUL_RESTART:
            from pbgpp.BGP.Open.Parameters.Capabilities.LongLivedGracefulRestart import CapabilityLongLivedGracefulRestart
            return CapabilityLongLivedGracefulRestart(payload)

        if capability_type == BGPStatics.CAPABILITY_FQDN:
            from pbgpp.BGP.Open.Parameters.Capabilities.FQDN import CapabilityFQDN
            return CapabilityFQDN(payload)

        # No type match
        from pbgpp.BGP.Open.Parameters.Capabilities.Unknown import CapabilityUnknown
        return CapabilityUnknown(payload, unknown_type=capability_type)



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import sys

from kafka import KafkaProducer
from kafka.errors import KafkaError

from pbgpp.BGP.Exceptions import BGPError
from pbgpp.Output.Pipe import BGPPipe


class KafkaPipe(BGPPipe):
    def __init__(self, server, topic):
        logger = logging.getLogger("pbgpp.KafkaPipe.__init__")

        # Kafka server initialization
        self.server = server
        self.topic = topic
        self.handle = None

        # Class specific variables
        try:
            self.handle = KafkaProducer(bootstrap_servers=[server])
        except Exception as e:
            logger.error("could not initialize connection to Apache Kafka server. Following exception has been reported: " + str(e))
            raise BGPError("Could not establish a connection to target pipe (Apache Kafka). Cancelling ...")

    def output(self, output):
        if self.handle is not None:
            if sys.version_info[0] < 3:
                self.handle.send(self.topic, output)
            else:
                self.handle.send(self.topic, bytes(output, "utf-8"))



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute


class PathAttributeLocalPreferences(BGPPathAttribute):
    def __init__(self, payload):
        BGPPathAttribute.__init__(self, payload)
        self.type = BGPStatics.UPDATE_ATTRIBUTE_LOCAL_PREF

        self.__parse()

    def __parse(self):
        self.parsed = True
        self.error = False



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.PathAttributes.ASPath import PathAttributeASPath
from pbgpp.BGP.Update.PathAttributes.Communities import PathAttributeCommunities
from pbgpp.BGP.Update.PathAttributes.LargeCommunities import PathAttributeLargeCommunities
from pbgpp.BGP.Update.PathAttributes.NextHop import PathAttributeNextHop
from pbgpp.BGP.Update.PathAttributes.Origin import PathAttributeOrigin
from pbgpp.Output.Formatter import BGPFormatter
from itertools import chain


class LineBasedFormatter(BGPFormatter):
    FIELD_MESSAGE_TIMESTAMP = ["timestamp"]
    FIELD_MESSAGE_IP_SOURCE = ["source_ip", "src_ip"]
    FIELD_MESSAGE_IP_DESTINATION = ["destination_ip", "dst_ip"]
    FIELD_MESSAGE_MAC_SOURCE = ["source_mac", "src_mac", "mac_src", "mac_source"]
    FIELD_MESSAGE_MAC_DESTINATION = ["destination_mac", "dst_mac", "mac_dst", "mac_destination"]
    FIELD_MESSAGE_LENGTH = ["length"]
    FIELD_MESSAGE_TYPE = ["type"]

    FIELD_UPDATE_SUBTYPE = ["subtype"]
    FIELD_UPDATE_PATH_ATTRIBUTES_LENGTH = ["path_attributes_length"]
    FIELD_UPDATE_WITHDRAWN_ROUTES_LENGTH = ["withdrawn_routes_length"]
    FIELD_UPDATE_WITHDRAWN_ROUTES = ["withdrawn_routes", "withdrawn_route", "withdrawals"]
    FIELD_UPDATE_NLRI = ["prefixes", "prefix", "nlri"]
    FIELD_UPDATE_NLRI_LENGTH = ["prefix_length"]
    FIELD_UPDATE_ATTRIBUTE_ORIGIN = ["origin"]
    FIELD_UPDATE_ATTRIBUTE_AS_PATH = ["as_path"]
    FIELD_UPDATE_ATTRIBUTE_AS_PATH_LAST_ASN = ["as_path_last_asn"]
    FIELD_UPDATE_ATTRIBUTE_NEXT_HOP = ["next_hop"]
    FIELD_UPDATE_ATTRIBUTE_COMMUNITIES = ["communities"]
    FIELD_UPDATE_ATTRIBUTE_LARGE_COMMUNITIES = ["large_communities"]

    FIELD_OPEN_MYASN = ["myasn", "my_asn", "asn"]
    FIELD_OPEN_HOLD_TIME = ["hold_time", "holdtime", "holdtimer", "hold_timer"]
    FIELD_OPEN_VERSION = ["version"]
    FIELD_OPEN_BGP_IDENTIFIER = ["bgp_identifier"]

    REGISTERED_FIELDS = [FIELD_MESSAGE_TIMESTAMP,
                         FIELD_MESSAGE_IP_SOURCE,
                         FIELD_MESSAGE_IP_DESTINATION,
                         FIELD_MESSAGE_MAC_SOURCE,
                         FIELD_MESSAGE_MAC_DESTINATION,
                         FIELD_MESSAGE_LENGTH,
                         FIELD_MESSAGE_TYPE,
                         FIELD_UPDATE_SUBTYPE,
                         FIELD_UPDATE_PATH_ATTRIBUTES_LENGTH,
                         FIELD_UPDATE_WITHDRAWN_ROUTES_LENGTH,
                         FIELD_UPDATE_WITHDRAWN_ROUTES,
                         FIELD_UPDATE_NLRI,
                         FIELD_UPDATE_NLRI_LENGTH,
                         FIELD_UPDATE_ATTRIBUTE_ORIGIN,
                         FIELD_UPDATE_ATTRIBUTE_AS_PATH,
                         FIELD_UPDATE_ATTRIBUTE_AS_PATH_LAST_ASN,
                         FIELD_UPDATE_ATTRIBUTE_NEXT_HOP,
                         FIELD_UPDATE_ATTRIBUTE_COMMUNITIES,
                         FIELD_UPDATE_ATTRIBUTE_LARGE_COMMUNITIES,
                         FIELD_OPEN_MYASN,
                         FIELD_OPEN_HOLD_TIME,
                         FIELD_OPEN_VERSION,
                         FIELD_OPEN_BGP_IDENTIFIER]

    def __init__(self, fields=None, separator='\t'):
        if not fields:
            # If the user is not using --fields parameter fallback to default field set
            # Default: timestamp, message_type, message_subtype, prefixes, withdrawn_routes
            self.fields = [self.FIELD_MESSAGE_TIMESTAMP, self.FIELD_MESSAGE_TYPE, self.FIELD_UPDATE_SUBTYPE, self.FIELD_UPDATE_NLRI, self.FIELD_UPDATE_WITHDRAWN_ROUTES]
        else:
            self.fields = fields

        self.separator = separator

    @staticmethod
    def available_fields():
        output = ""

        for field in LineBasedFormatter.REGISTERED_FIELDS:
            output += ", " + field[0]

        return output[2:]

    @staticmethod
    def is_registered(field):
        for f in LineBasedFormatter.REGISTERED_FIELDS:
            for item in f:
                if item == field:
                    return True

        return False

    def get_field_value(self, f, message):
        # Timestamp
        if f in self.FIELD_MESSAGE_TIMESTAMP:
            return str(message.pcap_information.get_timestamp()[0]) + "." + str(message.pcap_information.get_timestamp()[1])

        # Source IP
        if f in self.FIELD_MESSAGE_IP_SOURCE:
            return message.pcap_information.get_ip().get_source_string()

        # Destination IP
        if f in self.FIELD_MESSAGE_IP_DESTINATION:
            return message.pcap_information.get_ip().get_destination_string()

        # Source MAC
        if f in self.FIELD_MESSAGE_MAC_SOURCE:
            return message.pcap_information.get_mac().get_source_string()

        # Destination MAC
        if f in self.FIELD_MESSAGE_MAC_DESTINATION:
            return message.pcap_information.get_mac().get_destination_string()

        # ASN
        if f in self.FIELD_OPEN_MYASN:
            asn = getattr(message, "asn", False)
            if asn:
                return str(asn)
            return None

        # Hold time
        if f in self.FIELD_OPEN_HOLD_TIME:
            hold_time = getattr(message, "hold_time", False)
            if hold_time:
                return str(hold_time)
            return None

        # BGP Version
        if f in self.FIELD_OPEN_VERSION:
            version = getattr(message, "version", False)
            if version:
                return str(version)
            return None

        # BGP Identifier
        if f in self.FIELD_OPEN_BGP_IDENTIFIER:
            bgp_identifier = getattr(message, "identifier", False)
            if bgp_identifier:
                return str(bgp_identifier)
            return None

        # Message length
        if f in self.FIELD_MESSAGE_LENGTH:
            return str(message.length)

        # Path attributes length
        if f in self.FIELD_UPDATE_PATH_ATTRIBUTES_LENGTH:
            path_attributes_length = getattr(message, "path_attributes_length", False)
            if path_attributes_length:
                return str(path_attributes_length)
            return None

        # Withdrawn routes length
        if f in self.FIELD_UPDATE_WITHDRAWN_ROUTES_LENGTH:
            withdrawn_routes_length = getattr(message, "withdrawn_routes_length", False)
            if withdrawn_routes_length:
                return str(withdrawn_routes_length)
            return None

        # Message type
        if f in self.FIELD_MESSAGE_TYPE:
            return BGPTranslation.message_type(message.type)

        # Message sub type
        if f in self.FIELD_UPDATE_SUBTYPE:
            subtype = getattr(message, "subtype", False)
            if subtype:
                return BGPTranslation.update_subtype(subtype)
            return None

        # Withdrawn routes
        if f in self.FIELD_UPDATE_WITHDRAWN_ROUTES:
            w_routes = getattr(message, "withdrawn_routes", False)
            if w_routes:
                return [str(r) for r in w_routes]
            return None

        # NLRI (announced prefixes)
        if f in self.FIELD_UPDATE_NLRI:
            prefixes = getattr(message, "nlri", False)
            if prefixes:
                return [str(r) for r in prefixes]
            return None

        # NLRI length
        if f in self.FIELD_UPDATE_NLRI_LENGTH:
            prefixes = getattr(message, "nlri", False)
            if prefixes:
                return [r.prefix_length_string for r in prefixes]
            return None

        # Attribute: Origin
        if f in self.FIELD_UPDATE_ATTRIBUTE_ORIGIN:
            path_attributes = getattr(message, "path_attributes", False)
            if path_attributes:
                return [str(a) for a in path_attributes if isinstance(a, PathAttributeOrigin)]
            return None

        # Next hop
        if f in self.FIELD_UPDATE_ATTRIBUTE_NEXT_HOP:
            path_attributes = getattr(message, "path_attributes", False)
            if path_attributes:
                return [str(a) for a in path_attributes if isinstance(a, PathAttributeNextHop)]
            return None

        # Communities
        if f in self.FIELD_UPDATE_ATTRIBUTE_COMMUNITIES or f in self.FIELD_UPDATE_ATTRIBUTE_LARGE_COMMUNITIES:
            path_attributes = getattr(message, "path_attributes", False)
            if path_attributes:  # The split is necessary because communities are not returned as a list?
                communities = [str(pa).split(" ") for pa in path_attributes if isinstance(pa, PathAttributeCommunities)]
                return list(chain.from_iterable(communities))  # Flattens the resulting list
            return None

        # AS path (AS Sets remain unhandled?)
        if f in self.FIELD_UPDATE_ATTRIBUTE_AS_PATH:
            path_attributes = getattr(message, "path_attributes", False)
            if path_attributes:  # The split is necessary because segments are not returned as a list?
                segments = [str(seg).split(" ") for seg in path_attributes if isinstance(seg, PathAttributeASPath)]
                return list(chain.from_iterable(segments))  # Flattens the resulting list
            return None

    def apply(self, message):

        # Get the values
        values = [self.get_field_value(f, message) for f in self.fields]

        # Convert and return to a nice output string, assuming list of lists contained in values
        r = ""
        for i in values:
            if not isinstance(i, list):
                r += str(i)
            else:
                r += " ".join(map(str, i))
            r += self.separator
        return r[:-1]

        # # Convert to a nice output string
        # r = ""
        #
        # for f in self.fields:
        #     if f in self.FIELD_MESSAGE_TIMESTAMP:
        #         r += self.separator + str(message.pcap_information.get_timestamp()[0]) + "." + str(message.pcap_information.get_timestamp()[1])
        #     elif f in self.FIELD_OPEN_MYASN:
        #         # We can only display this information if we are handling an OPEN message
        #         if message.type == BGPStatics.MESSAGE_TYPE_OPEN:
        #             r += self.separator + str(message.asn)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_OPEN_HOLD_TIME:
        #         # We can only display this information if we are handling an OPEN message
        #         if message.type == BGPStatics.MESSAGE_TYPE_OPEN:
        #             r += self.separator + str(message.hold_time)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_OPEN_VERSION:
        #         # We can only display this information if we are handling an OPEN message
        #         if message.type == BGPStatics.MESSAGE_TYPE_OPEN:
        #             r = self.separator + str(message.version)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_OPEN_BGP_IDENTIFIER:
        #         # We can only display this information if we are handling an OPEN message
        #         if message.type == BGPStatics.MESSAGE_TYPE_OPEN:
        #             r += self.separator + str(message.identifier)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_MESSAGE_IP_SOURCE:
        #         r += self.separator + message.pcap_information.get_ip().get_source_string()
        #     elif f in self.FIELD_MESSAGE_IP_DESTINATION:
        #         r += self.separator + message.pcap_information.get_ip().get_destination_string()
        #     elif f in self.FIELD_MESSAGE_MAC_SOURCE:
        #         r += self.separator + message.pcap_information.get_mac().get_source_string()
        #     elif f in self.FIELD_MESSAGE_MAC_DESTINATION:
        #         r += self.separator + message.pcap_information.get_mac().get_destination_string()
        #     elif f in self.FIELD_MESSAGE_LENGTH:
        #         r += self.separator + str(message.length)
        #     elif f in self.FIELD_MESSAGE_TYPE:
        #         r += self.separator + BGPTranslation.message_type(message.type)
        #     elif f in self.FIELD_UPDATE_SUBTYPE:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             r += self.separator + BGPTranslation.update_subtype(message.subtype)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_PATH_ATTRIBUTES_LENGTH:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             r += self.separator + str(message.path_attributes_length)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_WITHDRAWN_ROUTES_LENGTH:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             r += self.separator + str(message.withdrawn_routes_length)
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_WITHDRAWN_ROUTES:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             if len(message.withdrawn_routes) > 0:
        #                 add = ""
        #
        #                 for route in message.withdrawn_routes:
        #                     add += ";" + str(route)
        #
        #                 # Skip first separator character
        #                 r += self.separator + add[1:]
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_NLRI:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             if len(message.nlri) > 0:
        #                 add = ""
        #
        #                 for route in message.nlri:
        #                     add += ";" + str(route)
        #
        #                 # Skip first separator character
        #                 r += self.separator + add[1:]
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_NLRI_LENGTH:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             if len(message.nlri) > 0:
        #                 add = ""
        #
        #                 for route in message.nlri:
        #                     add += ";" + str(route.prefix_length_string)
        #
        #                 # Skip first separator character
        #                 r += self.separator + add[1:]
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #
        #     elif f in self.FIELD_UPDATE_ATTRIBUTE_ORIGIN:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             if len(message.path_attributes) > 0:
        #                 for attribute in message.path_attributes:
        #                     # We found the correct path attribute
        #                     if isinstance(attribute, PathAttributeOrigin):
        #                         r += self.separator + str(attribute)
        #                         break
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #
        #     #elif f in self.FIELD_UPDATE_ATTRIBUTE_AS_PATH:
        #     #    print "Drinne"
        #
        #     elif True:
        #         print "F drin: ",
        #         print f in self.FIELD_UPDATE_ATTRIBUTE_AS_PATH
        #         if f in self.FIELD_UPDATE_ATTRIBUTE_AS_PATH:
        #             # We can only display this information if we are handling an UPDATE message
        #             if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #                 if len(message.path_attributes) > 0:
        #                     for attribute in message.path_attributes:
        #                         # We found the correct path attribute
        #                         if isinstance(attribute, PathAttributeASPath):
        #                             if len(attribute.path_segments) > 0:
        #                                 add = ""
        #
        #                                 for segment in attribute.path_segments:
        #                                     add += ";" + str(segment)
        #
        #                                 # Skip first separator
        #                                 r += self.separator + add[1:]
        #
        #                                 break
        #                             else:
        #                                 r += self.separator
        #                                 break
        #                 else:
        #                     r += self.separator
        #             else:
        #                 r += self.separator
        #
        #     elif f in self.FIELD_UPDATE_ATTRIBUTE_NEXT_HOP:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
        #             if len(message.path_attributes) > 0:
        #                 for attribute in message.path_attributes:
        #                     # We found the correct path attribute
        #                     if isinstance(attribute, PathAttributeNextHop):
        #                         r += self.separator + str(attribute)
        #                         nhf = True
        #                         break
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #
        #     elif f in self.FIELD_UPDATE_ATTRIBUTE_COMMUNITIES or f == self.FIELD_UPDATE_ATTRIBUTE_LARGE_COMMUNITIES:
        #         # We can only display this information if we are handling an UPDATE message
        #         if message.type == BGPStatics.MESSAGE_TYPE_UPDATE and message.subtype == BGPStatics.UPDATE_TYPE_ANNOUNCE:
        #             if len(message.path_attributes) > 0:
        #
        #                 communities = None
        #                 for attribute in message.path_attributes:
        #                     if isinstance(attribute, PathAttributeCommunities):
        #                         communities = attribute.communities
        #                     elif isinstance(attribute, PathAttributeLargeCommunities):
        #                         communities = attribute.large_communities
        #
        #                 if communities:
        #                     if len(communities) > 0:
        #                         add = ""
        #                         for community in communities:
        #                             add += ";" + str(community)
        #
        #                         r += self.separator + add[1:]
        #                         break
        #                     else:
        #                         r += self.separator
        #                 else:
        #                     r += self.separator
        #                     break
        #             else:
        #                 r += self.separator
        #         else:
        #             r += self.separator
        #     elif f in self.FIELD_UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES:
        #         # @todo Find a good way to display extended communities in just one line
        #         r += self.separator
        #     else:
        #         # No field match
        #         pass
        #
        # print r[1:]
        #
        # return r[1:]



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Exceptions import BGPOptionalParameterFactoryError
from pbgpp.BGP.Statics import BGPStatics


class BGPOptionalParameter:

    def __init__(self, payload):
        self.payload = payload

        self.type = None
        self.error = None
        self.parsed = False

    @staticmethod
    def factory(parameter_type, payload):
        # Factory pattern for the optional parameters of OPEN message
        if parameter_type == BGPStatics.OPEN_CAPABILITY:
            from pbgpp.BGP.Open.Parameters.Capability import BGPOptionalParameterCapability
            return BGPOptionalParameterCapability(payload)

        if parameter_type == BGPStatics.OPEN_AUTHENTICATION:
            from pbgpp.BGP.Open.Parameters.Authentication import BGPOptionalParameterAuthentication
            return BGPOptionalParameterAuthentication(payload)

        if parameter_type == BGPStatics.OPEN_RESERVED:
            from pbgpp.BGP.Open.Parameters.Reserved import BGPOptionalParameterReserved
            return BGPOptionalParameterReserved(payload)

        # No type match
        raise BGPOptionalParameterFactoryError("given parameter type is not valid")



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct
import logging
from binascii import hexlify

from pbgpp.PCAP.Information import PCAPLayer2Information


class PCAPEthernet:
    ETH_TYPE_IPV4 = 0x0800

    def __init__(self, payload):
        self.payload = payload
        self.type = None
        self.mac = None

        self.parsing_error = False
        self.parsed = False

        self.__parse()

    def __parse(self):
        try:
            self.parsed = True

            # Ethernet type
            self.type = struct.unpack("!H", self.payload[12:14])[0]

            # MAC addresses
            self.mac = PCAPLayer2Information(self.payload[6:12], self.payload[:6])

        except Exception as e:
            logging.error("Parsing ethernet frame caused exception (message: " + e.message + ")")
            self.parsing_error = True

    def get_type(self):
        return self.type

    def get_mac(self):
        return self.mac

    def get_payload(self):
        return self.payload

    def get_eth_payload(self):
        return self.payload[14:]

    def __str__(self):
        if self.parsed:
            if not self.parsing_error:
                # Preparing information
                type = hexlify(self.payload[12:14]).decode("utf-8")
                length = len(self)
                source = self.get_mac().get_source_string()
                destination = self.get_mac().get_destination_string()

                # and return formatted output
                return "<Ethernet type={0} length={1} source={2} destination={3}>".format(type, length, source, destination)
            else:
                logging.debug("Returning string representation of malformed ethernet frame")
                return "<MalformedEthernet length={0}>".format(len(self.payload))
        else:
            logging.debug("Returning string representation of not yet parsed ethernet frame")
            return "<Ethernet type=UNKNOWN length=UNKNOWN source=UNKNOWN destination=UNKNOWN>"

    def __len__(self):
        # len(obj) should represent the length of the payload
        if self.parsed and not self.parsing_error:
            return len(self.payload)
        else:
            logging.debug("Returning zero-length due to malformed or not yet parsed ethernet frame")
            return 0

    def __eq__(self, other):
        # Compare the exact byte payload to determine if an ethernet packet equals another one
        if self.payload == other.payload:
            return True
        else:
            return False



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from pbgpp.BGP.Update.Route import BGPRoute
from pbgpp.Output.Filter import BGPFilter
from pbgpp.BGP.Statics import BGPStatics


# @todo Verify functionality
class ASNFilter(BGPFilter):
    def __init__(self, values=[]):
        BGPFilter.__init__(self, values)

    def apply(self, message):
        try:
            # AS_PATH is a path attribute of BGP UPDATE message
            # Therefore we first need to make that we are currently handling an UPDATE message
            if message.type is not BGPStatics.MESSAGE_TYPE_UPDATE:
                # Skip messages that are no UPDATE messages
                return None

            for attribute in message.path_attributes:
                # Skip attributes that are no NEXT_HOP attributes
                if attribute.type is not BGPStatics.UPDATE_ATTRIBUTE_AS_PATH:
                    continue

                for path_segment in attribute.path_segments:
                    asn = list(map(str, path_segment.segments))

                    for value in self.values:
                        if value in asn:
                            return message

                        # Negative filtering using ~ character
                        if value[0:1] == "~" and value[1:] not in asn:
                            return message

            # Searched value was not found
            return None
        except Exception as e:
            # On error the filtering was not successful (due to wrong fields, etc.)
            return None



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import json

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.Output.Formatter import BGPFormatter


class JSONFormatter(BGPFormatter):
    def __init__(self):
        pass

    def apply(self, message):

        # Basic data for every message type
        data = {
            "timestamp": str(message.pcap_information.get_timestamp()[0]) + "." + str(message.pcap_information.get_timestamp()[1]),

            "message_type": message.type,
            "message_type_string": BGPTranslation.message_type(message.type),
            "length": message.length,

            "source_mac": message.pcap_information.get_mac().get_source_string(),
            "destination_mac": message.pcap_information.get_mac().get_destination_string(),
            "source_ip": message.pcap_information.get_ip().get_source_string(),
            "destination_ip": message.pcap_information.get_ip().get_destination_string(),

            "message_data": None
        }

        # Handle specific message types that contain more information than added above
        # Currently we just need to add information to OPEN- and UPDATE-messages
        if message.type == BGPStatics.MESSAGE_TYPE_OPEN:
            message_data = {
                "asn": message.asn,
                "hold_time": message.hold_time,
                "identifier": message.identifier,
                "optional_parameter_length": message.optional_parameter_length,

                "optional_parameters": None
            }

            # Add optional parameters
            optional_parameters = []

            if len(message.optional_parameter) > 0:
                for o in message.optional_parameter:
                    optional_parameters.append(o.json())

            # Assign to message data
            message_data["optional_parameters"] = optional_parameters

            # Assign message data to return data
            data["message_data"] = message_data

        elif message.type == BGPStatics.MESSAGE_TYPE_UPDATE:
            message_data = {
                "sub_type_string": BGPTranslation.update_subtype(message.subtype),
                "withdrawn_routes_length": message.withdrawn_routes_length,
                "path_attributes_length": message.path_attributes_length,

                "path_attributes": None,
                "withdrawn_routes": None,
                "nlri": None
            }

            path_attributes = []
            withdrawn_routes = []
            nlri = []

            # Add path attributes
            if len(message.path_attributes) > 0:
                for a in message.path_attributes:
                    path_attributes.append(a.json())

            # Add withdrawn routes
            if len(message.withdrawn_routes) > 0:
                for w in message.withdrawn_routes:
                    withdrawn_routes.append(str(w))

            # Add NLRI
            if len(message.nlri) > 0:
                for n in message.nlri:
                    nlri.append(str(n))

            # Assign to message data
            message_data["path_attributes"] = path_attributes
            message_data["withdrawn_routes"] = withdrawn_routes
            message_data["nlri"] = nlri

            # Assign message data to return data
            data["message_data"] = message_data

        return json.dumps(data)



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import struct

from pbgpp.BGP.Exceptions import BGPOptionalParameterFactoryError
from pbgpp.BGP.Message import BGPMessage
from pbgpp.BGP.Open.OptionalParameter import BGPOptionalParameter
from pbgpp.BGP.Statics import BGPStatics


class BGPOpenMessage(BGPMessage):
    def __init__(self, payload, length, pcap_information):
        BGPMessage.__init__(self, payload, length, pcap_information)
        self.type = BGPStatics.MESSAGE_TYPE_OPEN
        self.optional_parameter = []
        self.__parse()

    def __parse(self):
        self.parsed = True
        logger = logging.getLogger('pbgpp.BGPOpenMessage.__parse')

        try:
            fields = struct.unpack("!BHHLB", self.payload[:10])
            self.version = fields[0]
            self.asn = fields[1]
            self.hold_time = fields[2]
            self.identifier = fields[3]
            self.optional_parameter_length = fields[4]

            # Parse optional parameters if there are any
            if self.optional_parameter_length > 0:

                # First of all check if the message is malformed
                # The optional parameter length should equal the length of the remaining payload
                if self.optional_parameter_length == len(self.payload[10:]):
                    current_byte_position = 10  # Starting at 10 bytes
                    continue_loop = True

                    while continue_loop:
                        parameter_fields = struct.unpack("!BB", self.payload[current_byte_position:current_byte_position + 2])
                        parameter_type = parameter_fields[0]
                        parameter_length = parameter_fields[1]

                        parameter_payload_start = current_byte_position + 2  # Skip 2 additional bytes for the parameter type and length that were assigned beforehand
                        parameter_payload_stop = parameter_payload_start + parameter_length  # The payload of the optional parameter ends at (payload start marker + length of parameter)

                        # Now building the optional parameter factory
                        self.optional_parameter.append(BGPOptionalParameter.factory(parameter_type, self.payload[parameter_payload_start:parameter_payload_stop]))

                        # If adding was successful we want to continue the loop if there are more parameters left
                        # Therefore we need to adjust the current_byte_position
                        current_byte_position += (2 + parameter_length)  # New byte_position is (old position + (2 bytes header information + parameter length))

                        # If the current byte position minus 10 bytes offset from before is >= optional parameter length
                        if (current_byte_position - 10) >= self.optional_parameter_length:
                            continue_loop = False

                    # Parsing was successful
                    self.error = False

                else:
                    # Optional parameter length does NOT equal the length of the remaining payload - malformed message
                    logger.warning("Optional parameter length does not equal the length of remaining payload.")
                    self.error = True
            else:
                # Parsing was successful
                self.error = False

        except BGPOptionalParameterFactoryError as f:
            # Factory was not able to recognize parameter type - malformed message
            logger.warning("BGPOptionalParameterFactory was not able to recognize parameter type. Exception could be raised due to a malformed message.")
            self.error = True

        except Exception as e:
            # Other strange things could happen
            logger.warning("Unspecified error during packet parsing. Exception could be raised due to a malformed message.")



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from binascii import hexlify

import struct

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.ExtendedCommunity import BGPUpdateExtendedCommunity
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute


class PathAttributeExtendedCommunities(BGPPathAttribute):
    def __init__(self, payload, attribute_flags):
        BGPPathAttribute.__init__(self, payload)
        self.type = BGPStatics.UPDATE_ATTRIBUTE_EXTENDED_COMMUNITIES

        # We need the attribute flags for that path attribute
        self.attribute_flags = attribute_flags
        self.extended_communities = []

        self.__parse()

    def __parse(self):
        try:
            self.parsed = True
            self.error = False

            continue_loop = True
            current_byte_position = 0

            while continue_loop:

                # It's a lot easier and faster when we are checking types and sub-types like this
                fields = struct.unpack("!BBHI", self.payload[current_byte_position:current_byte_position + 8])
                current_byte_position += 8

                try:
                    # One extended community has always 8 bytes payload
                    self.extended_communities.append(BGPUpdateExtendedCommunity(fields[0], fields[1], fields[2], fields[3]))
                except Exception as e:
                    self.error = True

                if current_byte_position >= len(self.payload):
                    continue_loop = False

        except Exception as e:
            self.error = True

    def json(self):
        r = {
            "type": self.type,
            "type_string": BGPTranslation.path_attribute(self.type),
            "error": self.error,
            "extended_communities": []
        }

        for e in self.extended_communities:
            r["extended_communities"].append(e.json())

        return r



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import sys
import os.path
import glob
from itertools import chain

import pcapy

from pbgpp.BGP.Exceptions import BGPPacketHasNoMessagesError, BGPError
from pbgpp.BGP.Packet import BGPPacket
from pbgpp.Output.Filters.ASNFilter import ASNFilter
from pbgpp.Output.Filters.BlackholeFilter import BlackholeFilter
from pbgpp.Output.Filters.CommunityASNFilter import CommunityASNFilter
from pbgpp.Output.Filters.CommunityValueFilter import CommunityValueFilter
from pbgpp.Output.Filters.LargeCommunityFilter import LargeCommunityFilter
from pbgpp.Output.Filters.IPDestinationFilter import IPDestinationFilter
from pbgpp.Output.Filters.IPSourceFilter import IPSourceFilter
from pbgpp.Output.Filters.LastASNFilter import LastASNFilter
from pbgpp.Output.Filters.MACDestinationFilter import MACDestinationFilter
from pbgpp.Output.Filters.MACSourceFilter import MACSourceFilter
from pbgpp.Output.Filters.MessageSizeFilter import MessageSizeFilter
from pbgpp.Output.Filters.MessageSubTypeFilter import MessageSubTypeFilter
from pbgpp.Output.Filters.MessageTypeFilter import MessageTypeFilter
from pbgpp.Output.Filters.NLRIFilter import NLRIFilter
from pbgpp.Output.Filters.NextHopFilter import NextHopFilter
from pbgpp.Output.Filters.TimestampFilter import TimestampFilter
from pbgpp.Output.Filters.WithdrawnFilter import WithdrawnFilter
from pbgpp.Output.Formatters.HumanReadable import HumanReadableFormatter
from pbgpp.Output.Formatters.JSON import JSONFormatter
from pbgpp.Output.Formatters.LineBased import LineBasedFormatter
from pbgpp.Output.Handler import OutputHandler
from pbgpp.Output.Pipes.FilePipe import FilePipe
from pbgpp.Output.Pipes.KafkaPipe import KafkaPipe
from pbgpp.Output.Pipes.StdOutPipe import StdOutPipe
from pbgpp.PCAP.CookedCapture import PCAPCookedCapture
from pbgpp.PCAP.Ethernet import PCAPEthernet
from pbgpp.PCAP.IP import PCAPIP
from pbgpp.PCAP.Information import PCAPInformation
from pbgpp.PCAP.TCP import PCAPTCP


class PBGPPHandler:
    def __init__(self, parser):
        self.__parser = parser
        self.args = parser.parse_args()

        self.quiet = False
        self.verbose = False

        self.fields = None

        self.kafka_server = None
        self.kafka_topic = None

        self.formatter = None
        self.pipe = None
        self.filters = []
        self.prefilters = []

        self.__packet_counter = 0

    def handle(self):
        logger = logging.getLogger('pbgpp.PBGPPHandler.handle')

        if self.args.version:
            print("pbgpp PCAP BGP Parser v0.2.19")
            print("Copyright 2016-2018, DE-CIX Management GmbH")
            sys.exit(0)

        if self.args.quiet:
            # 60 is higher then CRITICAL log level
            logging.getLogger().setLevel(60)

        if self.args.verbose:
            logging.getLogger().setLevel(logging.DEBUG)

        logger.debug("Parsing filters ...")
        self.__parse_filters()

        logger.debug("Parsing formatters ...")
        self.__parse_formatter()

        logger.debug("Parsing pipes ...")
        self.__parse_pipe()

        # Check for input method
        if self.args.interface:
            logger.info("Initial startup finished. Calling interface handler ...")
            self.__handle_interface()
            logger.info("Parsing finished - Exiting now with status code 0")
            sys.exit(0)

        if self.args.pcap:
            logger.info("Initial startup finished. Calling pcap handler ...")
            self.__handle_pcap()
            logger.info("Parsing finished - Exiting now with status code 0")
            sys.exit(0)

        if self.args.stdin:
            logger.info("Initial startup finished. Calling stdin handler ...")
            self.__handle_stdin()
            logger.info("Parsing finished - Exiting now with status code 0")
            sys.exit(0)

        self.__parser.print_help()
        sys.exit(0)

    def __parse_filters(self):
        logger = logging.getLogger("pbgpp.PBGPPHandler.__parse_filters")

        if self.args.filter_message_type:
            values = self.args.filter_message_type
            filters = list(chain(*values))
            self.filters.append(MessageTypeFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of MessageTypeFilter")

        if self.args.filter_message_subtype:
            values = self.args.filter_message_subtype
            filters = list(chain(*values))
            self.filters.append(MessageSubTypeFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of MessageSubTypeFilter")

        if self.args.filter_nlri:
            values = self.args.filter_nlri
            filters = list(chain(*values))
            self.filters.append(NLRIFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of NLRIFilter")

        if self.args.filter_withdrawn:
            values = self.args.filter_withdrawn
            filters = list(chain(*values))
            self.filters.append(WithdrawnFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of WithdrawnFilter")

        if self.args.filter_next_hop:
            values = self.args.filter_next_hop
            filters = list(chain(*values))
            self.filters.append(NextHopFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of NextHopFilter")

        if self.args.filter_asn:
            values = self.args.filter_asn
            filters = list(chain(*values))
            self.filters.append(ASNFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of ASNFilter")

        if self.args.filter_last_asn:
            values = self.args.filter_last_asn
            filters = list(chain(*values))
            self.filters.append(LastASNFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of LastASNFilter")

        if self.args.filter_community_as:
            values = self.args.filter_community_as
            filters = list(chain(*values))
            self.filters.append(CommunityASNFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of CommunityASNFilter")

        if self.args.filter_community_value:
            values = self.args.filter_community_value
            filters = list(chain(*values))
            self.filters.append(CommunityValueFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of CommunityValueFilter")

        if self.args.filter_large_community:
            values = self.args.filter_large_community
            filters = list(chain(*values))
            self.filters.append(LargeCommunityFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of LargeCommunityFilter")

        if self.args.filter_message_size:
            values = self.args.filter_message_size
            filters = list(chain(*values))
            self.filters.append(MessageSizeFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter(s) of MessageSizeFilter")

        if self.args.filter_blackhole:
            values = self.args.filter_blackhole
            filters = list(chain(*values))
            self.filters.append(BlackholeFilter(filters))
            logger.debug("Added " + str(len(filters)) + " filter of BlackholeFilter")

        if self.args.filter_source_ip:
            values = self.args.filter_source_ip
            filters = list(chain(*values))
            self.prefilters.append(IPSourceFilter(filters))
            logger.debug("Added " + str(len(filters)) + " pre-filter(s) of IPSourceFilter")

        if self.args.filter_destination_ip:
            values = self.args.filter_destination_ip
            filters = list(chain(*values))
            self.prefilters.append(IPDestinationFilter(filters))
            logger.debug("Added " + str(len(filters)) + " pre-filter(s) of IPDestinationFilter")

        if self.args.filter_source_mac:
            values = self.args.filter_source_mac
            filters = list(chain(*values))
            self.prefilters.append(MACSourceFilter(MACSourceFilter.clear_input(filters)))
            logger.debug("Added " + str(len(filters)) + " pre-filter(s) of MACSourceFilter")

        if self.args.filter_destination_mac:
            values = self.args.filter_destination_mac
            filters = list(chain(*values))
            self.prefilters.append(MACDestinationFilter(MACSourceFilter.clear_input(filters)))
            logger.debug("Added " + str(len(filters)) + " pre-filter(s) of MACDestinationFilter")

        if self.args.filter_timestamp:
            values = self.args.filter_timestamp
            filters = list(chain(*values))
            self.prefilters.append(TimestampFilter(filters))
            logger.debug("Added " + str(len(filters)) + " pre-filter(s) of TimestampFilter")

    def __parse_formatter(self):
        if self.args.formatter == "JSON":
            self.formatter = JSONFormatter()
        elif self.args.formatter == "HUMAN_READABLE":
            self.formatter = HumanReadableFormatter()
        elif self.args.formatter == "LINE":
            values = self.args.fields.split(",")

            for v in values:
                if not LineBasedFormatter.is_registered(v):
                    self.__parser.error("Could not recognize field '" + str(v) + "' for line based output. Use --help argument to see all available fields.")

            self.formatter = LineBasedFormatter(fields=values)
        else:
            self.__parser.error("Can't recognize the formatter.")

    def __parse_pipe(self):
        if self.args.pipe == "FILE":
            if self.args.output_target is None:
                self.__parser.error("You need to specify the output target (-o / --output) when using FILE as pipe.")
            self.pipe = FilePipe(self.args.output_target)
        elif self.args.pipe == "STDOUT":
            self.pipe = StdOutPipe()
        elif self.args.pipe == "KAFKA":
            if self.args.kafka_server is None or self.args.kafka_topic is None:
                self.__parser.error("You need to specify Kafka server (--kafka-server) and topic (--kafka-topic) when using KAFKA as output pipe.")
            self.pipe = KafkaPipe(server=self.args.kafka_server, topic=self.args.kafka_topic)
        else:
            self.__parser.error("Can't recognize the output pipe.")

    def __handle_interface(self):
        # This is experimental! Not verified, yet.
        handle = pcapy.open_live(self.args.interface, 65536, 1, 0)
        handle.loop(0, self.__packet_handler)

    def __handle_pcap(self):
        logger = logging.getLogger("pbgpp.PBGPPHandler.__handle_pcap")

        if os.path.isfile(self.args.pcap):
            handle = pcapy.open_offline(self.args.pcap)
            handle.loop(0, self.__packet_handler)
        else:
            logger.info("Given PCAP input string is not direct path to a single file. Checking for glob-argument.")

            files = glob.glob(self.args.pcap)
            if len(files) == 0:
                logger.warning("Tried to use glob() on provided --pcap argument but list size is zero.")
                self.__parser.error("Specified --pcap argument is neither a single file nor a valid wildcard string (no files found!)")

            for f in files:
                logger.debug("Handling file: " + str(f))
                handle = pcapy.open_offline(f)
                handle.loop(0, self.__packet_handler)

    def __handle_stdin(self):
        handle = pcapy.open_offline("-")
        handle.loop(0, self.__packet_handler)

    def __packet_handler(self, header, payload):
        logger = logging.getLogger("pbgpp.PBGPPHandler.__packet_handler")
        logger.debug("Parsing PCAP packet " + str(self.__packet_counter))

        eth = PCAPEthernet(payload)

        # Check for raw ethernet packet
        if not eth.get_type() == PCAPEthernet.ETH_TYPE_IPV4:

            # Check for SLL-packet
            eth = PCAPCookedCapture(payload)

            if not eth.get_type() == PCAPCookedCapture.ETH_TYPE_IPV4:
                logger.debug("Discarding PCAP packet " + str(self.__packet_counter) + " due to non-IPv4 ethernet type.")
                return False

        ip = PCAPIP(eth.get_eth_payload())

        if not ip.get_protocol() == PCAPIP.PROTO_TCP:
            logger.debug("Discarding PCAP packet " + str(self.__packet_counter) + " due to non-TCP IP type.")
            return False

        tcp = PCAPTCP(ip.get_ip_payload())

        pcap_information = PCAPInformation(header.getts(), eth.mac, ip.addresses, tcp.ports)

        for filter in self.prefilters:
            if not filter.apply(pcap_information):
                logger.debug("Discarding PCAP packet " + str(self.__packet_counter) + " because no applied pre-filter could be matched.")
                return

        try:
            bgp = BGPPacket(tcp.get_tcp_payload(), pcap_information)

            messages = bgp.message_list

            for m in messages:
                handler = OutputHandler(message=m, filter=self.filters, formatter=self.formatter, pipe=self.pipe)
                handler.handle()

        except BGPPacketHasNoMessagesError:
            # This is no problem because a PCAP file could also contain TCP control packets. Those packets obviously do not contain any BGP information.
            logger.debug("BGPPacket which was assembled from PCAP packet " + str(self.__packet_counter) + " did not contain any BGP messages.")
        except BGPError:
            logger.error("Unspecified BGPError was raised while handling BGPPacket which was assembled from PCAP packet " + str(self.__packet_counter) + ".")
        finally:
            self.__packet_counter += 1



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPUpdateExtendedCommunity:
    def __init__(self, type, sub_type, global_administrator, local_administrator):
        self.type = type
        self.sub_type = sub_type
        self.global_administrator = global_administrator
        self.local_administrator = local_administrator

    def __str__(self):
        if isinstance(self.global_administrator, int) and isinstance(self.local_administrator, int) and isinstance(self.type, int) and isinstance(self.sub_type, int):
            type = self.type
            sub_type = self.sub_type
            global_administrator = self.global_administrator
            local_administrator = self.local_administrator

        elif isinstance(self.global_administrator, bytes) and isinstance(self.local_administrator, bytes) and isinstance(self.type, bytes) and isinstance(self.sub_type, bytes):
            type = struct.unpack("!B", self.type)[0]
            sub_type = struct.unpack("!B", self.sub_type)[0]
            global_administrator = struct.unpack("!H", self.global_administrator)[0]
            local_administrator = struct.unpack("!I", self.local_administrator)[0]

        else:
            return ""

        type_string = BGPTranslation.extended_community(type, sub_type)

        return type_string + " (" + str(self.global_administrator) + ":" + str(self.local_administrator) + ")"

    def json(self):
        r = {
            "type": self.type,
            "type_string": BGPTranslation.extended_community_type(self.type),
            "sub_type": self.sub_type,
            "sub_type_string": BGPTranslation.extended_community_subtype(self.type, self.sub_type),
            "global_administrator": self.global_administrator,
            "local_administrator": self.local_administrator
        }

        return r



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
import logging
import struct

from pbgpp.BGP.Exceptions import BGPMessageFactoryError
from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation


class BGPMessage:
    def __init__(self, payload, length, pcap_information):
        self.payload = payload
        self.length = length
        self.type = None
        self.parsed = False
        self.error = None
        self.pcap_information = pcap_information

    def __str__(self):
        # Return the string identifier of the BGP message
        # Example return: <BGPMessage type=UPDATE length=128>
        message_type_string = BGPTranslation.message_type(self.type)
        message_length_string = self.length
        message_parsed_string = BGPTranslation.boolean(self.parsed)
        message_error_string = BGPTranslation.boolean(self.error)

        # Assemble first part of message
        return_string = "<BGPMessage type="

        # Assemble message type
        if message_type_string is not None:
            return_string += message_type_string
        else:
            return_string += "UNKNOWN"

        # Assemble length
        return_string += " " + "length="

        if message_length_string is not None:
            return_string += str(message_length_string)
        else:
            return_string += "UNKNOWN"

        # Assemble parsed
        return_string += " " + "parsed="

        if message_parsed_string is not None:
            return_string += message_parsed_string
        else:
            return_string += "UNKNOWN"

        # Assemble error
        return_string += " " + "error="

        if message_error_string is not None:
            return_string += message_error_string
        else:
            return_string += "UNKNOWN"

        # Close message
        return_string += ">"

        # Return fully assembled message
        return return_string

    def __eq__(self, other):
        if isinstance(other, BGPMessage):
            # Compare raw payload to equate both objects
            return self.payload == other.payload
        else:
            return NotImplemented

    def __len__(self):
        # Return the message length
        return self.length

    def get_type(self):
        # Return the message type
        return self.type

    def get_length(self):
        # Return the message length
        return self.length

    @staticmethod
    def factory(payload, pcap_information):
        logger = logging.getLogger("pbgpp.BGPMessage.factory")

        # Implement factory pattern for easy message class creation
        # First 2 bytes of BGP header is the message length
        # The byte after message length is the message type
        try:
            bgp_header = struct.unpack("!HB", payload[:3])
        except Exception as e:
            # This could happen on a malformed packet
            logger.debug("Unpacking first 3 bytes of BGP message (length and type) failed.")
            raise BGPMessageFactoryError("given payload has no valid message type.")

        message_length = bgp_header[0]
        message_type = bgp_header[1]

        # Plausibility-check for BGP messages
        if message_length is not (len(payload) + 16):
            logger.warning("The unpacked message length does not equal the real payload length.")
            raise BGPMessageFactoryError("parsed message length does not equal real payload length.")

        if message_type == BGPStatics.MESSAGE_TYPE_UPDATE:
            from pbgpp.BGP.Update.Message import BGPUpdateMessage
            return BGPUpdateMessage(payload[3:], message_length, pcap_information)

        if message_type == BGPStatics.MESSAGE_TYPE_KEEPALIVE:
            from pbgpp.BGP.Keepalive.Message import BGPKeepaliveMessage
            return BGPKeepaliveMessage(payload[3:], message_length, pcap_information)

        if message_type == BGPStatics.MESSAGE_TYPE_OPEN:
            from pbgpp.BGP.Open.Message import BGPOpenMessage
            return BGPOpenMessage(payload[3:], message_length, pcap_information)

        if message_type == BGPStatics.MESSAGE_TYPE_NOTIFICATION:
            from pbgpp.BGP.Notification.Message import BGPNotificationMessage
            return BGPNotificationMessage(payload[3:], message_length, pcap_information)

        if message_type == BGPStatics.MESSAGE_TYPE_ROUTE_REFRESH:
            from pbgpp.BGP.RouteRefresh.Message import BGPRouteRefreshMessage
            return BGPRouteRefreshMessage(payload[3:], message_length, pcap_information)

        # No type match
        logger.warning("Factory could not recognize message type")
        raise BGPMessageFactoryError("given payload has no valid message type.")



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import logging
import struct

from pbgpp.BGP.Exceptions import BGPWithdrawnPrefixError, BGPNLRIError, BGPError
from pbgpp.BGP.Message import BGPMessage
from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Update.Flags import BGPUpdateFlags
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute
from pbgpp.BGP.Update.Route import BGPRoute


class BGPUpdateMessage(BGPMessage):
    def __init__(self, payload, length, pcap_information):
        BGPMessage.__init__(self, payload, length, pcap_information)
        self.type = BGPStatics.MESSAGE_TYPE_UPDATE
        self.subtype = BGPStatics.UPDATE_TYPE_NONE

        # Message specific variables
        self.path_attributes = []
        self.path_attributes_length = None

        self.withdrawn_routes = []
        self.withdrawn_routes_length = None

        self.nlri = []

        self.__parse()

    def __parse(self):
        self.parsed = True

        try:
            # Check last two bytes of byte payload - if they are set to zero we don't have any path attributes
            if struct.unpack("!H", self.payload[-2:])[0] == 0:
                self.path_attributes_length = 0

            # Unpack the length of withdrawn routes field and add 2 bytes to the current byte marker position
            self.withdrawn_routes_length = struct.unpack("!H", self.payload[:2])[0]
            current_byte_position = 2

            # Start parsing withdrawn routes
            if self.withdrawn_routes_length is not 0:
                continue_loop = True

                # Loop through withdrawals
                while continue_loop:
                    # First of all we need to parse the length of the withdrawn prefix. Depending on the prefix length
                    # we can determine the length following prefix itself
                    prefix_length_bytes = self.payload[current_byte_position:current_byte_position + 1]
                    prefix_length = struct.unpack("!B", prefix_length_bytes)[0]
                    current_byte_position += 1

                    if 0 <= prefix_length <= 8:
                        # Length of prefix field: 1 Byte
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 1]
                        current_byte_position += 1

                    elif 9 <= prefix_length <= 16:
                        # Length of prefix field: 2 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 2]
                        current_byte_position += 2

                    elif 17 <= prefix_length <= 24:
                        # Length of prefix field: 3 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 3]
                        current_byte_position += 3

                    elif 25 <= prefix_length:
                        # Length of prefix field: 4 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 4]
                        current_byte_position += 4

                    else:
                        self.error = True
                        raise BGPWithdrawnPrefixError("can't match prefix length.")

                    # Add BGPRoute object with information about the withdrawn route to list
                    self.withdrawn_routes.append(BGPRoute.from_binary(prefix_bytes, prefix_length_bytes))

                    # Check if we are at the end of the payload
                    if self.withdrawn_routes_length <= current_byte_position:
                        # If yes we need to stop the iteration
                        continue_loop = False

            # Second step: Continue with the path attributes
            if self.path_attributes_length is None:
                # First of all get the attributes length field and update the current byte position
                self.path_attributes_length = struct.unpack("!H", self.payload[current_byte_position:current_byte_position + 2])[0]
                current_byte_position += 2

            # Now we have a correct path_attributes_length stored. If this length is zero we don't need to do anything
            if self.path_attributes_length is not 0:
                continue_loop = True

                # Loop through path attributes
                while continue_loop:
                    # Now comes a tricky part of UPDATE message parsing. Each path attribute has a flag bitfield.
                    # One of those flags is called 'extended length'. If it's set to 1 the following attribute fields
                    # are 2 bytes long. But if it's set to zero it's just 1 byte long ...

                    # So first of all: Flag parsing!
                    attribute_flags = BGPUpdateFlags(struct.unpack("!B", self.payload[current_byte_position:current_byte_position + 1])[0])
                    current_byte_position += 1

                    if attribute_flags.length:
                        # We got an extended length flag
                        attribute_fields = struct.unpack("!BH", self.payload[current_byte_position:current_byte_position + 3])
                        current_byte_position += 3
                    else:
                        # We got a normal length flag
                        attribute_fields = struct.unpack("!BB", self.payload[current_byte_position:current_byte_position + 2])
                        current_byte_position += 2

                    # Finally assign the variables
                    attribute_type = attribute_fields[0]
                    attribute_length = attribute_fields[1]

                    # Now we are using the factory pattern again to determine
                    # which kind of attribute we have to add the list
                    self.path_attributes.append(BGPPathAttribute.factory(attribute_type, self.payload[current_byte_position:current_byte_position + attribute_length], attribute_flags))

                    # Add length of attribute to position pointer
                    current_byte_position += attribute_length

                    # Check if there are further path attributes to parse
                    if current_byte_position >= self.path_attributes_length:
                        continue_loop = False

            # Third step: NLRIs
            if len(self.payload) > (self.path_attributes_length + 4 + self.withdrawn_routes_length):

                continue_loop = True

                # The four bytes that get added to the attributes and routes length are the two two-byte fields
                # for path attribute length and withdrawn routes length
                current_byte_position = self.path_attributes_length + 4 + self.withdrawn_routes_length

                while continue_loop:
                    # First of all we have to check the prefix length as byte-length of the following
                    # prefix depends on its prefix length (This is a 1-byte-field)
                    prefix_length_bytes = self.payload[current_byte_position:current_byte_position + 1]
                    prefix_length = struct.unpack("!B", prefix_length_bytes)[0]
                    current_byte_position += 1

                    if 0 <= prefix_length <= 8:
                        # Length of prefix field: 1 Byte
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 1]
                        current_byte_position += 1
                    elif 9 <= prefix_length <= 16:
                        # Length of prefix field: 2 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 2]
                        current_byte_position += 2
                    elif 17 <= prefix_length <= 24:
                        # Length of prefix field: 3 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 3]
                        current_byte_position += 3
                    elif 25 <= prefix_length:
                        # Length of prefix field: 4 Bytes
                        prefix_bytes = self.payload[current_byte_position:current_byte_position + 4]
                        current_byte_position += 4
                    else:
                        self.error = True
                        raise BGPWithdrawnPrefixError("can't match prefix length.")

                    try:
                        self.nlri.append(BGPRoute.from_binary(prefix_bytes, prefix_length_bytes))
                    except BGPError as e:
                        raise BGPNLRIError("can't append NLRI to message (error: " + str(e) + ")")

                    if current_byte_position >= len(self.payload):
                        continue_loop = False

            # Determine sub-type for filtering
            # Using bit flags for easier assignment
            if len(self.nlri) > 0:
                self.subtype = (self.subtype | BGPStatics.UPDATE_TYPE_ANNOUNCE)

            if len(self.withdrawn_routes) > 0:
                self.subtype = (self.subtype | BGPStatics.UPDATE_TYPE_WITHDRAWAL)

        except BGPWithdrawnPrefixError as p:
            self.error = True
            logging.info(p)
        except Exception as e:
            self.error = True

        self.error = False



from setuptools import setup, find_packages

packages = ['pbgpp.' + pkg for pkg in find_packages('pbgpp')]
packages.append('pbgpp')

setup(
    name='pbgpp',
    version='0.2.19',
    description='PCAP BGP Parser',
    author='DE-CIX Management GmbH',
    author_email='rnd@de-cix.net',
    url='https://github.com/de-cix/pbgp-parser',
    keywords=['bgp', 'parsing', 'pcap'],
    license='Apache License 2.0',

    zip_safe=False,
    packages=packages,
    install_requires=['pcapy', 'kafka-python'],
    entry_points={
        'console_scripts': [
            'pbgpp = pbgpp.Application.CLI:main'
        ]
    }
)



#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import struct

from pbgpp.BGP.Statics import BGPStatics
from pbgpp.BGP.Translation import BGPTranslation
from pbgpp.BGP.Update.LargeCommunity import BGPUpdateLargeCommunity
from pbgpp.BGP.Update.PathAttribute import BGPPathAttribute


class PathAttributeLargeCommunities(BGPPathAttribute):
    def __init__(self, payload):
        BGPPathAttribute.__init__(self, payload)
        self.type = BGPStatics.UPDATE_ATTRIBUTE_LARGE_COMMUNITIES

        # Path attribute specific variable
        self.large_communities = []

        self.__parse()

    def __str__(self):
        return_string = ""
        first = True

        for community in self.large_communities:
            if first:
                first = False
                return_string += str(community)
                continue

            return_string += " " + str(community)

        return None if len(return_string) == 0 else return_string

    def __parse(self):
        try:
            self.parsed = True
            self.error = False

            community_count = len(self.payload) / 12
            current_count = 1
            current_byte_position = 0

            while current_count <= community_count:
                fields = struct.unpack("!LLL", self.payload[current_byte_position:current_byte_position + 12])
                global_administrator = fields[0]
                local_data_part_1 = fields[1]
                local_data_part_2 = fields[2]
                self.large_communities.append(BGPUpdateLargeCommunity(global_administrator, local_data_part_1, local_data_part_2))

                current_count += 1
                current_byte_position += 12

        except Exception as e:
            self.error = True

    def json(self):
        r = {
            "type": self.type,
            "type_string": BGPTranslation.path_attribute(self.type),
            "error": self.error,
            "large_communities": []
        }

        for community in self.large_communities:
            r["large_communities"].append(community.json())

        return r



#!/usr/bin/env python
#
# This file is part of PCAP BGP Parser (pbgpp)
#
# Copyright 2016-2017 DE-CIX Management GmbH
# Author: Tobias Hannaske <tobias.hannaske@de-cix.net>
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

import argparse
import logging

import sys

from pbgpp.Application.Handler import PBGPPHandler
from pbgpp.Output.Formatters.LineBased import LineBasedFormatter


def main():
    logger = logging.getLogger('pbgpp')
    logging.basicConfig(level=logging.INFO, format="%(asctime)s - %(name)s - %(levelname)s - %(message)s")
    logging.getLogger().setLevel(logging.ERROR)

    parser = argparse.ArgumentParser(description="detailed bgp packet message parsing from PCAP files or direct network traffic")

    parser.add_argument("-f", "--formatter", help="specify data output format", choices=['JSON', 'HUMAN_READABLE', 'LINE'], default="HUMAN_READABLE", dest="formatter")
    parser.add_argument("-p", "--pipe", help="specify output target type of parsed messages", choices=['FILE', 'STDOUT', 'KAFKA'], default="STDOUT", dest="pipe")
    parser.add_argument("-o", "--output", help="specify target output file if your output type is set to FILE", dest="output_target")

    group_1 = parser.add_mutually_exclusive_group()
    group_1.add_argument("--interface", help="use a network interface as input  (specify interface)", dest="interface")
    group_1.add_argument("--pcap", help="use a pcap file as input (specify file)", dest="pcap")
    group_1.add_argument("--stdin", "-", help="use stdin as input", dest="stdin", action="store_true")

    group_2 = parser.add_mutually_exclusive_group()
    group_2.add_argument("-q", "--quiet", help="only show parsing output", action="store_true", dest="quiet")
    group_2.add_argument("-v", "--verbose", help="increase output verbosity", action="store_true", dest="verbose")

    group_3 = parser.add_argument_group("kafka configuration")
    group_3.add_argument("--kafka-server", help="IP address / hostname (and port if it's different from 9092) of the target Apache Kafka server if your output type is set to KAFKA (e.g. 127.0.0.1:9092)", dest="kafka_server")
    group_3.add_argument("--kafka-topic", help="topic of Apache Kafka server if your output type is set to KAFKA (e.g. pbgpp)", dest="kafka_topic")

    group_4 = parser.add_argument_group("filters")
    group_4.add_argument("--filter-timestamp", help="only print messages with given epoch timestamp (e.g., 123456789)", nargs="+", action="append", dest="filter_timestamp")
    group_4.add_argument("--filter-message-size", help="only print messages with given message size in bytes (e.g., 128)", nargs="+", action="append", dest="filter_message_size")
    group_4.add_argument("--filter-message-type", help="only print messages with given BGP message type (KEEPALIVE, NOTIFICATION, OPEN, ROUTE-REFRESH, UPDATE, WITHDRAWAL)", nargs="+", action="append", dest="filter_message_type")
    group_4.add_argument("--filter-message-subtype", help="only print UPDATE messages with given message sub type (WITHDRAWAL, ANNOUNCE, BOTH, NONE)", nargs="+", action="append", dest="filter_message_subtype")
    group_4.add_argument("--filter-nlri", help="only print messages containing the given nlri prefix (e.g., '80.81.82.0/24'", nargs="+", action="append", dest="filter_nlri")
    group_4.add_argument("--filter-withdrawn", help="only print messages containing the given withdrawn routes (e.g., '80.81.82.0/24'", nargs="+", action="append", dest="filter_withdrawn")
    group_4.add_argument("--filter-next-hop", help="only print messages containing the given next hop (e.g., '80.81.82.83')", nargs="+", action="append", dest="filter_next_hop")
    group_4.add_argument("--filter-as", help="only print messages containing the given ASN in path AS_PATH attribute (e.g., '12345')", nargs="+", action="append", dest="filter_asn")
    group_4.add_argument("--filter-last-as", help="only print messages containing the given ASN as last ASN in AS_PATH attribute (e.g., '12345')", nargs="+", action="append", dest="filter_last_asn")
    group_4.add_argument("--filter-community-as", help="only print messages containing the given community ASN (e.g., '12345')", nargs="+", action="append", dest="filter_community_as")
    group_4.add_argument("--filter-community-value", help="only print messages containing the given community value (e.g., '12345')", nargs="+", action="append", dest="filter_community_value")
    group_4.add_argument("--filter-source-ip", help="only print messages containing the given source IP address (e.g., '80.81.82.83')", nargs="+", action="append", dest="filter_source_ip")
    group_4.add_argument("--filter-source-mac", help="only print messages containing the given source MAC address (e.g., 'aabbccddeeff')", nargs="+", action="append", dest="filter_source_mac")
    group_4.add_argument("--filter-destination-ip", help="only print messages containing the given destination IP address (e.g., '80.81.82.83')", nargs="+", action="append", dest="filter_destination_ip")
    group_4.add_argument("--filter-destination-mac", help="only print messages containing the given destination MAC address (e.g., 'aabbccddeeff')", nargs="+", action="append", dest="filter_destination_mac")
    group_4.add_argument("--filter-large-community", help="only print messages containing one or more matching large communities (e.g., '11:22:33', '11:*:*', '*:22:33')", nargs="+", action="append", dest="filter_large_community")
    group_4.add_argument("--filter-blackhole", help="only print messages that contain blackhole prefixes with given next_hop (e.g. 80.81.193.66) OR RFC7999 well-known BGP community value", nargs="+", action="append", dest="filter_blackhole")

    group_5 = parser.add_argument_group("line output commands")
    group_5.add_argument("--fields", help="specify the output-fields to be display in the order desired; separated by comma. Available fields are: " + LineBasedFormatter.available_fields(), dest="fields", default=LineBasedFormatter.FIELD_MESSAGE_TIMESTAMP[0] + "," + LineBasedFormatter.FIELD_MESSAGE_TYPE[0] + "," + LineBasedFormatter.FIELD_UPDATE_SUBTYPE[0] + "," + LineBasedFormatter.FIELD_UPDATE_NLRI[0] + "," + LineBasedFormatter.FIELD_UPDATE_WITHDRAWN_ROUTES[0])

    group_6 = parser.add_argument_group("other commands")
    group_6.add_argument("--version", help="displays the current version of this software", action="store_true", dest="version")

    main_handler = PBGPPHandler(parser)

    try:
        main_handler.handle()
    except Exception as e:
        logger.error("Main error handler has received an exception: " + str(e))
        sys.exit(1)
    except KeyboardInterrupt:
        logger.info("Received KeyboardInterrupt - terminating ...")
        print("Exit execution due to keyboard interruption.")
        sys.exit(0)
